import java.io.StringReader
import org.apache.spark.{SparkContext, SparkConf}
import org.apache.spark.mllib.feature.{Word2Vec, Word2VecModel}
import org.apache.lucene.analysis.ja.JapaneseTokenizer
import org.apache.lucene.analysis.tokenattributes.CharTermAttribute
import org.apache.lucene.analysis.ja.tokenattributes.BaseFormAttribute
import org.apache.lucene.analysis.ja.tokenattributes.PartOfSpeechAttribute
import org.apache.spark.mllib.linalg.Vectors

object KoreanWord2VecExample {

  def tokenize(sentence: String) :Seq[String] = {

    // 단어 단위로 분할 된 임시저장배열
    val word: scala.collection.mutable.ArrayBuffer[String] = new scala.collection.mutable.ArrayBuffer[String]()
    // 문서의 토큰  스트림을 얻는다.
    lazy val stream = new JapaneseTokenizer(new StringReader(sentence), null, false, JapaneseTokenizer.Mode.NORMAL)

    try {
      // 스트림으로 부터 토큰을 하나식 얻어내 품사기준으로 분류
      while(stream.incrementToken()) {
        var charAtt = stream.getAttribute(classOf[CharTermAttribute]).toString
        var bfAtt = stream.getAttribute(classOf[BaseFormAttribute]).getBaseForm
        // 품사 하나를 끄집어 낸다
        var partOfSpeech = stream.getAttribute(classOf[PartOfSpeechAttribute]).getPartOfSpeech().split("-")(0)

        // 품사와 원형을 가지고 비교한다. 동사의 경우에는 원형을 등록한다.
        (partOfSpeech, bfAtt) match {
          case ("명사", _)    => word += charAtt
          case ("동사", null) => word += charAtt
          case ("동사", baseForm)    => word += baseForm
          case (_, _)         =>
        }
      }
    } finally {
      stream.close
    }

    word.toSeq
  }

  def relationWords(w1: String, w2: String,target: String,model: Word2VecModel) :Array[(String, Double)] = {
    // Breeze라는 Scala의 행렬계산을 할 수 있는 라이브러리를 이용한다.
    val b = breeze.linalg.Vector(model.getVectors(w1))
    val a = breeze.linalg.Vector(model.getVectors(w2))
    val c = breeze.linalg.Vector(model.getVectors(target))

    // v3에 대해서 v2 - v1 의 관계를 적용한다
    val x = c + (a - b)

    // t와 유사한 단어를 찾는다.
    model.findSynonyms(Vectors.dense(x.toArray.map(_.toDouble)),10)
  }

  def main(args: Array[String]) {

    if(args.length != 2) {
      System.err.println(
        "Usage: JapaneseWord2VecExample <inputdir> <outputdir>")
      System.exit(1)
    }

    val sparkConf = new SparkConf().setAppName("JapaneseWord2VecExample")
    val sc = new SparkContext(sparkConf)

    val input = sc.textFile(args(0)).map(line => tokenize(line))

    val word2vec = new Word2Vec()
    word2vec.setMinCount(3)
    word2vec.setVectorSize(30)
    val model = word2vec.fit(input)

    // 특정단어와 비슷한 단어를 10건 추출한다 (입력 데이터에 맞추어 적절히 변경해준다)
//    val word = "단어A"
//    println("Input word: " + word)
//    model.findSynonyms(word,10).foreach{
//      w => println("word:" + w._1 + "  score:" + w._2)
//    }

    // 단어A,B 의 관계를 단어C에 적용한다 (입력 데이터에 맞추어 적절히 변경해준다)
//    val wordA = "단어A"
//    val wordB = "단어B"
//    val wordC = "단어C"
//    println(wordC + " + ( " + wordA + " - " + wordB + " ) = ???")
//    relationWords(wordA,wordB,wordC,model).foreach{
//      w => println("word:" + w._1 + "  score:" + w._2)
//    }

    // 모델을 저장한다.
    model.save(sc, args(1))

  }
}
